#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/06/18


from pyxllib.prog.pupil import check_install_package

check_install_package('croniter')

import datetime
import time
import sys

from croniter import croniter


class SchedulerUtils:
    @classmethod
    def calculate_future_time(cls, start_time, wait_seconds):
        """ 计算延迟时间

        :param datetime start_time: 开始时间
        :param int wait_seconds: 等待秒数
            todo 先只支持秒数这种标准秒数，后续可以考虑支持更多智能的"1小时"等这种解析
        """
        return start_time + datetime.timedelta(seconds=wait_seconds)

    @classmethod
    def calculate_next_cron_time(cls, cron_tag, base_time=None):
        """ 使用crontab标记的运行周期，然后计算相对当前时间，下一次要启动运行的时间

        :param str cron_tag:
            30 2 * * 1: 这部分是时间和日期的设定，具体含义如下：
                30: 表示分钟，即每小时的第 30 分钟。
                2: 表示小时，即凌晨 2 点。
                第三个星号 *: 表示日，这里的星号意味着每天。
                第四个星号 *: 表示月份，星号同样表示每个月。
                1: 表示星期中的日子，这里的 1 代表星期一。
        """

        # 如果没有提供基准时间，则使用当前时间
        if base_time is None:
            base_time = datetime.datetime.now()
        # 初始化 croniter 对象
        cron = croniter(cron_tag, base_time)
        # 计算下一次运行时间
        next_time = cron.get_next(datetime.datetime)
        return next_time

    @classmethod
    def wait_until_time(cls, dst_time):
        """
        :param datetime dst_time: 一直等待到目标时间
            期间可以用time.sleep进行等待
        """
        # 一般来说，只要计算一轮待等待秒数就行。但是time.sleep机制好像不一定准确的，所以使用无限循环重试会更好。
        while True:
            # 先计算当前时间和目标时间的相差秒数
            wait_seconds = (dst_time - datetime.datetime.now()).total_seconds()
            if wait_seconds <= 0:
                break
            time.sleep(max(1, wait_seconds))  # 最少等待1秒

    @classmethod
    def smart_wait(cls, start_time, end_time, wait_tag, print_mode=0):
        """ 智能等待，一般用在对进程的管理重启上

        :param datetime start_time: 程序启动的时间
        :param datetime end_time: 程序结束的时间
        :param str|float|int wait_tag: 等待标记
            如果是str，则按照crontab解析
            如果是int，则表示等待的秒数
            如果是float，则表示等待的秒数
        """
        # 1 尝试把wait_tag转成数值
        try:
            wait_tag = float(wait_tag)
        except ValueError:  # 转不成也没关系
            pass

        if start_time is None:
            start_time = datetime.datetime.now()
        if end_time is None:
            end_time = datetime.datetime.now()

        # 2 计算下一次启动时间
        if isinstance(wait_tag, str):
            # 按照crontab解析
            next_time = cls.calculate_next_cron_time(wait_tag, end_time)
        elif wait_tag >= 0:
            # 正值则是从end_time开始往后等待
            next_time = cls.calculate_future_time(end_time, wait_tag)
        elif wait_tag < 0:
            # 负值则是从start_time开始往前等待
            next_time = cls.calculate_future_time(start_time, wait_tag)
        else:
            raise ValueError

        if print_mode:
            print(f'等待到时间{next_time}...')

        cls.wait_until_time(next_time)


def trial():
    # 设置基准时间
    base_time = datetime.datetime.now()
    # 定义crontab表达式
    cron_expression = '30 2 * * 1'  # 每周一凌晨2点30分执行
    # 测试延迟功能
    delayed_time = SchedulerUtils.calculate_future_time(base_time, 10)  # 延迟10秒
    # 测试crontab计算下一次运行时间
    next_run_time = SchedulerUtils.calculate_next_cron_time(cron_expression, base_time)

    print(base_time, delayed_time, next_run_time, sep='\n')


if __name__ == '__main__':
    import fire

    # 1 如果输入命令行参数，使用fire机制运行
    if len(sys.argv) > 1:
        fire.Fire()
        exit()

    # 2 否则执行这里的测试代码
    trial()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/04/12 08:59

import copy
import json
import re
import sqlite3
import warnings
from collections import defaultdict

import pandas as pd

# 旧版的pandas警告
warnings.filterwarnings('ignore', message="pandas only support SQLAlchemy connectable")
# 新版的pandas警告多了个's‘
warnings.filterwarnings('ignore', message="pandas only supports SQLAlchemy connectable")


class SqlBuilder:
    def __init__(self, table=''):
        self.table = table
        self._select = []
        self._set = []
        self._where = []
        self._order_by = []
        self._group_by = []
        self._limit = None  # 限制最多读多少条数据
        self._offset = None  # 从第几条数据开始读

    def copy(self):
        # 拷贝一个当前状态的副本sql
        return copy.deepcopy(self)

    def __1_组件(self):
        pass

    def from_table(self, table):
        self.table = table
        return self

    def select(self, *columns):
        self._select.extend(columns)
        return self

    def set(self, *columns):
        self._set.extend(columns)
        return self

    def where(self, condition):
        if isinstance(condition, (list, tuple)):
            self._where.extend(condition)
        elif isinstance(condition, str):
            self._where.append(condition)
        else:
            raise ValueError(f'不支持的where条件类型{type(condition)}')

        return self

    def where_dict_match(self, items):
        """ 输入一个字典，要求每个key的字段对应上具体的某个取值value """
        for k, v in items.items():
            if isinstance(v, str):
                self._where.append(f"{k} = '{v}'")
            elif isinstance(v, (int, float)):
                self._where.append(f"{k} = {v}")
            else:
                raise TypeError

    def where_in(self, column, values):
        if values is None:
            return self

        if isinstance(values, str):
            values = [values]
        values_str = ', '.join(f"'{str(value)}'" for value in values)
        if len(values) == 1:
            self._where.append(f"{column} = {values_str[0]}")
        else:
            self._where.append(f"{column} IN ({values_str})")
        return self

    def where_or(self, *conditions):
        """ 输入的这一批条件，作为OR组合后成为一个整体条件
        """
        self._where.append(f"({' OR '.join(conditions)})")
        return self

    def where_mod(self, column, divisor, remainder):
        """ 输入的column列的值对divisor取余，筛选出余数为remainder的记录
        """
        condition = f"({column} % {divisor} = {remainder})"
        self._where.append(condition)
        return self

    def where_mod2(self, desc):
        """ 使用一种特殊的格式化标记来设置规则

        :param desc: 例如 'id%2=1'

        todo 我好像傻了，画蛇添足，这个功能好像直接用 .where 就能完成~
        """
        if not desc:
            return
        column, divisor_remainder = desc.split('%')
        divisor, remainder = map(int, divisor_remainder.split('='))
        return self.where_mod(column, divisor, remainder)

    def group_by(self, *columns):
        self._group_by.extend(columns)
        return self

    def order_by(self, *columns):
        self._order_by.extend(columns)
        return self

    def limit(self, limit, offset=None):
        self._limit = limit
        self._offset = offset
        return self

    def __2_build_初级命令(self):
        pass

    def build_select(self, *columns):
        if columns:
            columns = self._select + list(columns)
        else:
            columns = self._select

        sql = [f"SELECT {', '.join(columns) or '*'}", f"FROM {self.table}"]
        if self._where:
            sql.append(f"WHERE {' AND '.join(self._where)}")
        if self._group_by:
            sql.append(f"GROUP BY {', '.join(self._group_by)}")
        if self._order_by:
            sql.append(f"ORDER BY {', '.join(self._order_by)}")
        if self._limit is not None:
            limit_clause = f"LIMIT {self._limit}"
            if self._offset is not None:
                limit_clause += f" OFFSET {self._offset}"
            sql.append(limit_clause)
        return '\n'.join(sql)

    def build_count(self):
        sql = [f"SELECT COUNT(*)", f"FROM {self.table}"]
        if self._where:
            sql.append(f"WHERE {' AND '.join(self._where)}")
        if self._group_by:
            sql.append(f"GROUP BY {', '.join(self._group_by)}")
        return '\n'.join(sql)

    def build_update(self):
        sql = [f"UPDATE {self.table}"]
        if self._set:
            sql.append(f"SET {', '.join(self._set)}")
        if self._where:
            sql.append(f"WHERE {' AND '.join(self._where)}")
        return '\n'.join(sql)

    def __3_build_中级命令(self):
        pass

    def build_check_data_type(self, column):
        """ 检查column的数据类型 """
        sql = SqlBuilder('information_schema.columns')
        sql.select(f"data_type")
        sql.where(f"table_name='{self.table}' AND column_name='{column}'")
        return sql.build_select()

    def build_group_count(self, columns, count_column_name='cnt'):
        sql = SqlBuilder(self.table)
        if isinstance(columns, (list, tuple)):
            columns = ', '.join(columns)
        sql.select(columns, f"COUNT(*) {count_column_name}")
        sql.group_by(columns)
        sql.order_by(f'{count_column_name} DESC')
        sql._where = self._where.copy()
        return sql.build_select()


class SqlBase:
    """ Sql语法通用的功能 """

    def __init__(self, *args, **kwargs):
        self._commit_cache = defaultdict(list)

    def __1_库(self):
        pass

    def __2_表格(self):
        pass

    def get_table_names(self):
        """ 获得所有表格名 """
        raise NotImplementedError

    def create_table(self, table_name, column_descs):
        """ 【DDL增】

        :param table_name:
        :param column_descs:
            str, 正常的列格式描述，例如 'c1 text, c2 blob'
            dict, k是列名，v是一个具体的值，供分析参考格式类型
        :return:
        """
        # 1 列数据格式智能分析
        if not isinstance(column_descs, str):
            descs = []
            for k, v in column_descs.items():
                t = self.autotype(v)
                descs.append(f'{k} {t}')
            column_descs = ','.join(descs)

        # 2 新建表格
        self.execute(f'CREATE TABLE {table_name}({column_descs})')

    def get_column_names(self, table_name):
        """ 获得一个表格中所有字段名 """
        raise NotImplementedError

    def ensure_column(self, table_name, col_name, col_type='', *, col_ref_val=None, **kwargs):
        """ 【DDL改】添加字段

        :param table_name:
        :param col_name:
        :param col_type:
        :param col_ref_val: 可以通过具体的数值，智能判断列数据格式
        :return:
        """
        if col_name in self.get_column_names(table_name):
            return

        if col_ref_val:
            col_type = self.autotype(col_ref_val)
        if not col_type:
            col_type = 'text'

        self.execute(f'ALTER TABLE {table_name} ADD COLUMN {col_name} {col_type}')

    def delete_column(self, table_name, col_name):
        """ 删除表格中的一个列

        其实直接删除也不麻烦：ALTER TABLE {table_name} DROP COLUMN {col_name}
        这个封装主要是鲁棒性考虑，只有列存在时才会删除
        """
        if col_name in self.get_column_names(table_name):
            self.execute(f'ALTER TABLE {table_name} DROP COLUMN {col_name}')

    def create_index(self, index_name, table_name, cols):
        if not isinstance(cols, str):
            cols = ','.join(map(str, cols))
        self.execute(f'CREATE INDEX {index_name} ON {table_name}({cols})')

    def create_index2(self, table_name, cols):
        """ 创建一个简单的索引，索引名字自动生成 """
        if not isinstance(cols, str):
            cols = ','.join(map(str, cols))
        self.execute(f'CREATE INDEX idx_{table_name}_{cols.replace(",", "_")} ON {table_name}({cols})')

    def keep_top_n_rows(self, table_name, num, col_name='id'):
        """ 只保留一小部分数据，常用来做lite、demo数据示例文件

        :param col_name: 参照的列名
        """
        self.execute(f'DELETE FROM {table_name} WHERE {col_name} NOT IN'
                     f'(SELECT {col_name} FROM {table_name} LIMIT {num})')
        self.commit()

    def clear_table(self, table_name):
        """ 【DDL删】清空表格内容 """
        self.execute(f'DELETE FROM {table_name}')

    def count_all_talbe_rows(self):
        """ 统计所有表格的数据行数 """
        names = self.get_table_names()
        ls = []
        for name in names:
            n = self.execute(f'SELECT count(*) FROM {name}').fetchone()[0]
            ls.append([name, n])
        ls.sort(key=lambda x: -x[1])
        return ls

    def __3_execute(self):
        pass

    def exec2one(self, *args, **kwargs):
        """ 获得第1行的值 """
        try:
            return self.execute(*args, **kwargs).fetchone()[0]
        except TypeError:
            return None

    def exec2row(self, *args, **kwargs):
        """ 获得第1行的值 """
        return self.execute(*args, **kwargs).fetchone()

    def exec2col(self, *args, **kwargs):
        """ 获得第1列的值 """
        return [row[0] for row in self.execute(*args, **kwargs).fetchall()]

    # 兼容旧接口
    exec_col = exec2col

    def exec2df(self, *args, **kwargs):
        """ 获得pandas.DataFrame类型的返回值 """
        return pd.read_sql(*args, self, **kwargs)

    def __4_数据类型(self):
        pass

    @classmethod
    def cvt_type(cls, val):
        """ py一些内存对象，需要进行适当的格式转换，才能存储到sql中
        """
        raise NotImplementedError

    @classmethod
    def cvt_types(cls, vals):
        """ 批量转换类型
        """
        return [cls.cvt_type(v) for v in vals]

    @classmethod
    def autotype(cls, val):
        """ 自动判断一个py内存对象应该以什么类型进行存储 """
        raise NotImplementedError

    def __5_增删改查(self):
        pass

    def commit_base(self, commit_type, query, params=None):
        """
        :param commit_type:
            -1，先真正缓存在本地
            False，传统的事务机制，虽然不会更新数据，但每一条依然会连接数据库，其实速度回挺慢的
            True，传统的事务机制，但每条都作为独立事务，直接更新了
        """
        if commit_type == -1:
            self._commit_cache[query].append(params)
        elif commit_type is False:
            self.execute(query, params)
        elif commit_type is True:
            self.execute(query, params)
            self.commit()

    def commit_all(self):
        if not self._commit_cache:
            self.commit()
            return

        for query, params in self._commit_cache.items():
            cur = self.cursor()
            cur.executemany(query, params)
            cur.close()
            self.commit()

        self._commit_cache = defaultdict(list)

    def update_row(self, table_name, cols, where, *, commit=False):
        """ 【改】更新数据

        虽然名称是update_row，但where条件满足时，是有可能批量替换多行的

        :param dict cols: 要更新的字段及值
        :param dict where: 怎么匹配到对应记录
        :param commit: 建议减小commit频率，会极大降低性能
        :return:

        >> xldb.update('xlapi', {'input': d}, {'id': x['id']})
        """
        kvs = ','.join([f'{k}=%s' for k in cols.keys()])
        ops = ' AND '.join([f'{k}=%s' for k in where.keys()])
        vals = list(cols.values()) + list(where.values())
        self.commit_base(commit,
                         f'UPDATE {table_name} SET {kvs} WHERE {ops}',
                         self.cvt_types(vals))

    def delete_row(self, table_name, where, *, commit=False):
        """ 【删】删除数据

        :param dict where: 怎么匹配到对应记录
        :param commit: 建议减小commit频率，会极大降低性能
        :return:
        """
        ops = ' AND '.join([f'{k}=%s' for k in where.keys()])
        vals = list(where.values())
        self.commit_base(commit,
                         f'DELETE FROM {table_name} WHERE {ops}',
                         self.cvt_types(vals))

    def select_col(self, table_name, col):
        """ 获得一列数据，常使用的功能，所以做了一个封装

        注意，"exec"前缀的方法一般返回的是迭代器，而"select"前缀获得一般是直接的全部列表、结果
        """
        return self.exec2col(f'SELECT {col} FROM {table_name}')

    def group_count(self, table_name, cols, count_column_name='cnt'):
        """ 【查】分组统计各组值组合出现次数
        分析{table}表中，{cols}出现的种类和次数，按照出现次数从多到少排序

        :param str|list cols: 输入逗号','隔开的字符串，比如
            con.group_count('gpu_trace', 'host_name,total_memory')
            后记：写list也行，会自动join为字符串
        """
        if not isinstance(cols, str):
            cols = ','.join(map(str, cols))
        sql = f'SELECT {cols}, COUNT(*) {count_column_name} FROM {table_name} ' \
              f'GROUP BY {cols} ORDER BY {count_column_name} DESC'
        records = self.execute(sql).fetchall()
        df = pd.DataFrame.from_records(records, columns=cols.split(',') + [count_column_name])
        return df

    def get_count_by_altering_query(self, data_query: str) -> int:
        """
        从给定的SQL SELECT查询中获取行数计数。这个方法通过修改原始的SELECT查询，
        将其转换为一个COUNT查询来实现计数。这种方法特别适用于在获取大量数据之前，
        需要预估数据量的场景。

        问题背景：
        在进行大规模数据处理前，了解数据的规模可以帮助进行更有效的资源分配和性能优化。
        传统的做法是分两步执行：首先计算数据总量，然后再执行实际的数据提取。
        这个函数旨在通过单个查询来简化这一流程，减少数据库的负载和响应时间。

        实现机制：
        函数首先使用正则表达式识别出SQL查询的FROM关键词，这是因为无论SELECT查询的复杂程度如何，
        计数的核心都是保留FROM及其后面的表和条件语句。然后，它构造一个新的COUNT查询，
        替换原始查询中的SELECT部分。最后，函数执行这个新的查询并返回结果。

        :param data_query (str): 原始的SQL SELECT查询字符串。
        :return int: 查询结果的行数。

        示例:
        >> sql = SqlBase()
        >> query = "SELECT id, name FROM users WHERE active = True"
        >> count = sql.get_count_by_altering_query(query)
        >> print(count)
        45

        注意:
        - 这个函数假设输入的是合法的SQL SELECT查询。
        - 函数依赖于数据库连接的execute方法能够正确执行转换后的COUNT查询。
        - 在一些复杂的SQL查询中，特别是包含子查询、特殊函数或复杂的JOIN操作时，
          请确保转换后的计数查询仍然有效。
        """
        # 使用正则表达式定位'FROM'（考虑各种大小写情况），并确保它前后是空格或语句的开始/结束
        match = re.search(r'\bFROM\b', data_query, flags=re.IGNORECASE)
        if match:
            from_index = match.start()
            count_query = 'SELECT COUNT(*) ' + data_query[from_index:]  # 构造计数查询
            try:
                result = self.execute(count_query).fetchone()  # 执行查询
                return result[0] if result else 0  # 返回计数结果
            except Exception as e:
                print(f"Error executing count query: {e}")
                return 0
        else:
            print("No 'FROM' keyword found in the data query.")
            return 0

    def get_column_data_type(self, table_name, col_name):
        """ 获取表格中某一列的数据类型 """
        return self.exec2one(SqlBuilder(table_name).build_check_data_type(col_name))


class Connection(sqlite3.Connection, SqlBase):
    """
    DDL - 数据定义语言
        CREATE	创建一个新的表，一个表的视图，或者数据库中的其他对象。
        ALTER	修改数据库中的某个已有的数据库对象，比如一个表。
        DROP	删除整个表，或者表的视图，或者数据库中的其他对象。
    DML - 数据操作语言
        INSERT	创建一条记录。
        UPDATE	修改记录。
        DELETE	删除记录。
    DQL - 数据查询语言
        SELECT	从一个或多个表中检索某些记录。
    """

    def __1_库(self):
        pass

    def vacuum(self):
        """ 删除数据后，文件不会直接减小，需要使用vacuum来实际压缩文件占用空间 """
        self.execute('vacuum')  # 不用 commit

    def __2_表格(self):
        pass

    def get_table_names(self):
        """ 获得所有表格名 """
        return [x[0] for x in self.execute("SELECT name FROM sqlite_master WHERE type='table'")]

    def has_table(self, table_name):
        res = self.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'").fetchone()
        return bool(res)

    def get_column_names(self, table_name):
        """ 【查】表格有哪些字段
        """
        names = [item[1] for item in self.execute(f'PRAGMA table_info({table_name})')]
        return names

    def __3_execute(self):
        pass

    def exec2nametuple(self, *args, **kwargs):
        cur = self.cursor()
        cur.row_factory = sqlite3.Row
        return cur.execute(*args, **kwargs)

    def exec2dict(self, *args, **kwargs):
        """ execute基础上，改成返回值为dict类型 """

        def dict_factory(cursor, row):
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]
            return d

        cur = self.cursor()  # todo 不关是不是不好？如果出错了是不是会事务未结束导致无法修改表格结构？是否有auto close的机制？
        cur.row_factory = dict_factory
        return cur.execute(*args, **kwargs)

    # 兼容老版本
    exec_dict = exec2dict

    def __4_数据类型(self):
        pass

    @classmethod
    def cvt_type(cls, val):
        if isinstance(val, (dict, list, tuple)):
            val = json.dumps(val, ensure_ascii=False)
        return val

    @classmethod
    def autotype(cls, val):
        if isinstance(val, str):
            return 'text'
        elif isinstance(val, (int, bool)):
            return 'integer'
        elif isinstance(val, float):
            return 'real'
        else:  # 其他dict、list等类型，可以用json.dumps或str转文本存储
            return 'text'

    def __5_增删改查(self):
        pass

    def insert_row(self, table_name, cols, if_exists='IGNORE'):
        """ 【增】插入新数据

        :param table_name:
        :param cols: 一般是用字典表示的要插入的值
        :param if_exists: 如果已存在的处理策略
            IGNORE，跳过
            REPLACE，替换
        :return:

        注意加了 OR IGNORE，支持重复数据自动忽略插入
        """
        ks = ','.join(cols.keys())
        vs = ','.join('?' * (len(cols.keys())))
        self.execute(f'INSERT OR {if_exists} INTO {table_name}({ks}) VALUES ({vs})', self.cvt_types(cols.values()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/03/28

""" 使用subprocess来实现"多进程"并发 """

import os
import subprocess
import sys
import time
import atexit

from pyxllib.prog.pupil import tprint


class MultiProcessLauncher:
    def __init__(self):
        self.workers = []
        atexit.register(self.cleanup)  # 在程序退出时注册清理方法

    def _add_posix_process(self, cmd):
        import ctypes
        import signal

        def _set_pdeathsig(sig=signal.SIGTERM):
            """ Help function to ensure once parent process exits,
            its children processes will automatically die on Linux.
            """

            def callable():
                libc = ctypes.CDLL("libc.so.6")
                return libc.prctl(1, sig)

            return callable

        preexec_fn = _set_pdeathsig(signal.SIGTERM)
        p = subprocess.Popen(cmd.split(), preexec_fn=preexec_fn)

        return p

    def add_process(self, cmd, name=None, **kwargs):
        # 240329周五16:57，最新测试，好像不捕捉，主进程结束后子进程也是会自动结束的。跟我以前了解的机制似乎有点不一样。
        #   但加着也算双重保险，不会出错吧。
        if os.name == 'posix':
            p = self._add_posix_process(cmd)
        else:
            p = subprocess.Popen(cmd)

        worker = {
            'name': cmd.split()[0] if name is None else name,
            'process': p,
            'cmd': cmd,
            'status': 'running'  # 初始状态设为running
        }
        worker.update(kwargs)

        self.workers.append(worker)

    def add_cur_py_process(self, cmd, name=None, **kwargs):
        """ 添加当前python环境下的进程 """
        if isinstance(cmd, str):
            cmd = f'"{sys.executable}" {cmd}'
        elif isinstance(cmd, list):
            cmd = [sys.executable] + cmd

        self.add_process(cmd, name, **kwargs)

    def manage(self):
        """ 进入交互管理界面 """

        while True:
            running_count = 0
            for worker in self.workers:
                if worker['process'].poll() is None:  # 如果进程仍然在运行
                    running_count += 1
                else:
                    worker['status'] = 'completed'

            # 检查并输出状态
            if running_count == 0:
                print("所有进程都已完成。")
                break

            cmd = input(f'>>>  ')
            if cmd == "status":
                print(f"正在运行的进程数：{running_count}/{len(self.workers)}")
            # 你可以在这里根据需要扩展其他指令，比如停止特定进程等

    def wait_all(self):
        """ 进入交互管理界面 """

        seconds = 1  # 逐步增加等待间隔
        while True:
            running_count = 0
            for worker in self.workers:
                if worker['process'].poll() is None:  # 如果进程仍然在运行
                    running_count += 1
                else:
                    worker['status'] = 'completed'
                    tprint(f"{worker['name']} 已完成。")

            # 检查并输出状态
            if running_count == 0:
                print("所有进程都已完成。")
                break

            time.sleep(seconds)
            seconds = min(seconds + 1, 10)

    def cleanup(self):
        """ 在退出时终止所有子进程 """
        for worker in self.workers:
            process = worker['process']
            if process.poll() is None:  # 如果进程还在运行
                try:
                    process.terminate()  # 尝试终止进程
                    process.wait()  # 等待进程结束
                except Exception as e:
                    print(f"Error terminating process {worker['name']}: {e}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/04/12
import time

from pyxllib.prog.pupil import check_install_package

check_install_package('DrissionPage')

import re
from urllib.parse import unquote

import DrissionPage
from DrissionPage import ChromiumPage
from DrissionPage._pages.chromium_base import ChromiumBase
from DrissionPage._pages.chromium_tab import ChromiumTab
from DrissionPage._base.base import BasePage, BaseElement

from pyxllib.prog.pupil import inject_members
from pyxllib.text.pupil import strfind
from pyxllib.file.specialist import GetEtag


def get_dp_page(dp_page=None) -> 'XlPage':
    """

    :param dp_page:
        默认None, 返回默认的page，一般就是当前页面
        True, 新建一个page
        str, 新建一个对应url的page
        func(tab), 通过规则筛选tab，返回符合条件的第1个tab，否则新建一个tab
    """

    if isinstance(dp_page, ChromiumPage):
        return dp_page
    elif isinstance(dp_page, ChromiumTab):
        return dp_page.page
    elif callable(dp_page):
        page0 = ChromiumPage()
        for tab in page0.get_tabs():
            if dp_page(tab):
                return tab.page
        return page0.new_tab().page
    elif dp_page is True:
        return ChromiumPage().new_tab().page
    elif isinstance(dp_page, str):
        return ChromiumPage().new_tab(dp_page).page
    else:
        return ChromiumPage()


def get_dp_tab(dp_page=None) -> 'XlPage':
    """ 智能获取一个标签页tab

    :param dp_page:
        默认None, 返回默认的page，一般就是当前页面
        True, 新建一个page
        str, 新建一个对应url的page
        func(tab), 通过规则筛选tab，返回符合条件的第1个tab，否则新建一个tab
    """

    if isinstance(dp_page, ChromiumPage):
        return dp_page.latest_tab
    elif isinstance(dp_page, ChromiumTab):
        return dp_page
    elif callable(dp_page):
        page0 = ChromiumPage()
        for tab in page0.get_tabs():
            if dp_page(tab):
                return tab
        return page0.new_tab()
    elif dp_page is True:
        return ChromiumPage().new_tab()
    elif isinstance(dp_page, str):
        def dp_page2(tab):  # 默认开启页面复用。 todo 以后可以改进成可选参数
            if tab.url.startswith(dp_page):
                return tab

        return get_dp_page(dp_page2)
    else:
        return ChromiumPage().latest_tab


def get_latest_not_dev_tab(page=None):
    """ 开发工具本身也会算一个tab，这个函数返回最新的一个不是开发工具的tab """
    if page is None:
        page = ChromiumPage()
    tabs = page.get_tabs()
    for tab in tabs:
        if strfind(tab.url, ['devtools://', 'chrome-extension://']) != -1:
            continue
        return tab


def set_input_text(input_ele, text):
    """ 因为input输入框可能原本就自带了内容，为了不重复输入，先清空再输入 """
    input_ele.clear()
    input_ele.input(text)


def search_download_file(file_name):
    file_name = file_name.replace(':', '_')
    files = ChromiumPage().get_download_files()
    for file in files:
        if file_name in file['file']:  # 正常情况下的匹配
            return file
    for file in files:
        file2 = file['file'].replace('+', ' ')
        if file_name in file2:  # 但有时候'+'好像有点特别
            return file
        if file_name in re.sub(r'\s+', ' ', file2):
            return file


class XlChromiumBase(ChromiumBase):
    def get2(self, url, show_errmsg=False, retry=None, interval=None):
        """
        240418周四21:57，DrissionPage-4.0.4.21 官方自带page.get，有时候会有bug，不会实际刷新url，这里加个代码进行fix
        """
        old_url = self.url
        if old_url == url:
            # page.refresh()
            return

        ele = self.active_ele
        self.get(url, show_errmsg=show_errmsg, retry=retry, interval=interval)
        try:  # 如果新页面获取成功，理论上旧的ele会失效
            ele
            self.refresh()  # 如果不报错，这里网站要强制更新
            return self
        except DrissionPage.errors.ElementLostError:
            return self

    def get_download_files(self: ChromiumPage):
        """ 获取下载列表

        :param search_name: 搜索文件名，输入该参数时，只会从上往下找到第一个匹配的文件
            否则返回一个list结构，存储下载清单里的文件
        :return:

        todo 默认应该显示是不全的，有时间可以考虑往下滑动继续检索的功能
        """
        files = []
        page2 = self.new_tab('chrome://downloads/')
        items = page2('tag:downloads-manager', timeout=1).sr('#mainContainer')('#downloadsList').eles(
            'tag:downloads-item')
        for item in items:
            loc = unquote(item.sr('tag:img').attr('src').replace('+', ' '))
            file = re.search(r'path=(.+?)(&scale=(\d+(\.\d+)?)x)?$', loc).group(1)

            files.append({
                'file': file,
                'url': unquote(item.sr('#url').attr('href'))
            })

        page2.close()

        return files

    def wait_page_not_change(self, interval=3):
        """ 等待直到页面内容不再变化

        :param interval: 时间间隔，需要判断当前内容和interval秒后的内容，看内容是否欧发生改变
        """
        last_html, last_etag = None, None
        while True:
            html = self.html
            etag = GetEtag.from_text(html)
            if etag == last_etag:
                break

            last_html, last_etag = html, etag
            time.sleep(interval)
        return last_html

    def action_type(self, ele, text, clear=True):
        """ 基于action实现的重写入，常用于日期相关操作
        因为很多日期类组件，直接使用ele.input是不生效的，哪怕看似显示了文本，但其实并没有触发js改动，需要用动作链来实现
        """
        from DrissionPage.common import Keys
        if clear:
            self.actions.click(ele).key_down(Keys.CTRL).type('a').key_up(Keys.CTRL).type(text)
        else:
            self.actions.click(ele).type(text)


inject_members(XlChromiumBase, ChromiumBase)


class XlPage(XlChromiumBase, ChromiumTab, ChromiumPage):
    """ 只作为一个类型标记，无实质功能。在猴子补丁背景下，让ide能正确跳转函数定义。 """
    pass


def wait_page_not_change(page, interval=3):
    page.wait_page_not_change(interval)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/01/01


import requests

from pyxllib.xl import *
from pyxllib.algo.stat import *
from pprint import pprint

from pyxllib.text.nestenv import NestEnv


class Yuque:
    """
    https://www.yuque.com/yuque/developer/openapi
    语雀请求限制：每小时最多 5000 次请求，每秒最多 100 次请求
    """

    def __init__(self, token, user_id=None):
        self.base_url = "https://www.yuque.com/api/v2"
        self.headers = {
            "X-Auth-Token": token,
            "Content-Type": "application/json"
        }
        self._user_id = user_id

    def get_user(self):
        """ 获取用户信息

        # 获得的内容
         {'data':
            {'_serializer': 'v2.user',  # 数据序列化版本
              # 用户头像URL
              'avatar_url': 'https://cdn.nlark.com/yuque/0/2020/.../d5f8391e-2fdd-4f1b-8ea5-299be8fceecd.png',
              'books_count': 12,  # 知识库数量
              'created_at': '2018-11-16T07:26:27.000Z',  # 账户创建时间
              'description': '',  # 用户描述
              'followers_count': 54,  # 跟随者数量
              'following_count': 6,  # 关注者数量
              'id': 123456,  # 用户唯一标识
              'login': 'code4101',  # 用户登录名
              'name': '代号4101',  # 户昵称或姓名
              'public': 1,  # 用户公开状态，1为公开
              'public_books_count': 2,  # 公开的知识库数量
              'type': 'User',  # 数据类型，这里为'User'
              'updated_at': '2023-12-31T02:43:16.000Z',  # 信息最后更新时间
              'work_id': ''}}
        """
        url = f"{self.base_url}/user"
        response = requests.get(url, headers=self.headers)
        return response.json()

    @property
    def user_id(self):
        """ 很多接口需要用到用户ID，这里缓存一下 """
        if self._user_id is None:
            self._user_id = self.get_user()['data']['id']
        return self._user_id

    def __1_知识库操作(self):
        pass

    @run_once('id,str')  # todo 应该有更好的缓存机制，目前这样的实现，需要重启程序才会刷新
    def get_repos(self, return_mode=0):
        """ 获取某个用户的知识库列表

        :param int|str return_mode: 返回模式
            0（默认），返回原始json结构
            df，df结构
            nickname2id，获取知识库namespace或昵称到ID的映射
        """
        if return_mode == 0:
            url = f"{self.base_url}/users/{self.user_id}/repos"
            response = requests.get(url, headers=self.headers)
            return response.json()
        elif return_mode == 'df':
            data = self.get_repos()
            columns = ['id', 'name', 'items_count', 'namespace']

            ls = []
            for d in data['data']:
                ls.append([d[col] for col in columns])

            df = pd.DataFrame(ls, columns=columns)
            return df
        elif return_mode == 'nickname2id':  # namespace、name到id的映射（注意这里不考虑）
            data = self.get_repos()
            names2id = {d['name']: d['id'] for d in data['data']}
            # 例如："日志"知识库的namespace是journal，然后 journal -> 24363220
            namespace2id = {d['namespace'].split('/')[-1]: d['id'] for d in data['data']}
            names2id.update(namespace2id)
            return names2id
        else:
            raise ValueError(f'不支持的return_mode={return_mode}')

    def get_repo_id(self, repo_id):
        """ repo_id支持输入"昵称"来获得实际id
        """
        if isinstance(repo_id, str) and not re.match(r'\d+$', repo_id):
            repo_id = self.get_repos('nickname2id')[repo_id]
        return repo_id

    def get_repo_docs(self, repo_id, *, return_mode=0):
        """ 获取知识库的文档列表

        :param repo_id: 知识库的ID或Namespace（如"日志"是我改成的"journal"）
        :param int|str return_mode: 返回模式
            0（默认），返回原始json结构
            -1（df），df结构
        :return: 文档列表，只能获得最近的100篇文档
        """
        repo_id = self.get_repo_id(repo_id)
        if return_mode == 0:
            url = f"{self.base_url}/repos/{repo_id}/docs"
            response = requests.get(url, headers=self.headers)
            return response.json()
        elif return_mode in (-1, 'df'):
            data = self.get_repo_docs(repo_id)
            # 按照updated_at降序
            data['data'].sort(key=lambda x: x['updated_at'], reverse=True)
            columns = ['id', 'title', 'word_count', 'description', 'updated_at']

            ls = []
            for d in data['data']:
                ls.append([d.get(col) for col in columns])

            # ls.sort(key=lambda x: x[0])  # id一般就是创建顺序
            df = pd.DataFrame(ls, columns=columns)
            # df['updated_at']把'2024-08-07T06:13:10.000Z'转成datetime，并改到utf8时区
            df['updated_at'] = pd.to_datetime(df['updated_at']).dt.tz_convert('Asia/Shanghai')
            # 不显示时区
            df['updated_at'] = df['updated_at'].dt.strftime('%Y-%m-%d %H:%M:%S')
            return df
        else:
            raise ValueError(f'不支持的return_mode={return_mode}')

    def __2_文档操作(self):
        pass

    def create_doc(self, repo_id, title, md_content):
        """ 创建单篇文档，并放到目录开头
        """
        # 1 创建文档
        repo_id = self.get_repo_id(repo_id)
        url = f"{self.base_url}/repos/{repo_id}/docs"
        in_data = {
            "title": title,
            "body": md_content,
        }
        resp = requests.post(url, json=in_data, headers=self.headers)
        out_data = resp.json()
        doc_id = out_data['data']['id']

        # 2 将文档添加到目录中
        url2 = f"{self.base_url}/repos/{repo_id}/toc"
        in_data2 = {
            "action": "prependNode",  # 默认添加到知识库目录最顶上的位置
            "action_mode": "child",
            "doc_id": doc_id,
        }
        resp2 = requests.put(url2, json=in_data2, headers=self.headers)
        out_data2 = resp2.json()

        return out_data2

    def get_doc(self, repo_id, doc_id):
        """ 获取单篇文档的详细信息

        :param repo_id: 知识库的ID或Namespace
        :param doc_id: 文档的ID
        :return: 文档的详细信息
        """
        repo_id = self.get_repo_id(repo_id)
        url = f"{self.base_url}/repos/{repo_id}/docs/{doc_id}"
        response = requests.get(url, headers=self.headers)
        return response.json()

    def get_doc_from_url(self, url, return_mode='md'):
        """ 从文档的URL中获取文档的详细信息

        :param url: 文档的URL
        :param return_mode: 返回模式，
            json, 为原始json结构
            md, 返回文档的主体md内容
            title_and_md, 返回文档的标题和md内容
        :return: 文档的详细信息
        """
        repo_slug, doc_slug = url.split('/')[-2:]
        res = self.get_doc(repo_slug, doc_slug)

        if return_mode == 'json':
            return res['data']
        elif return_mode == 'md':
            return res['data']['body']
        elif return_mode == 'title_and_md':
            return f'文章标题：{res["data"]["title"]}\n\n{res["data"]["body"]}'

    def export_markdown(self, url, output_dir=None, post_mode=1):
        """ 导出md格式文件

        :param str|list[str] url: 文档的URL
            可以导出单篇文档，也可以打包批量导出多篇文档的md文件
        :param output_dir: 导出目录
            单篇的文件名是按照文章标题自动生成的
            多篇的可以自己指定具体文件名
        :param post_mode: 后处理模式
            0，不做处理
            1，做适当的精简
        """
        # 1 获得内容
        data = self.get_doc_from_url(url, return_mode='json')
        body = data['body']
        if post_mode == 0:
            pass
        elif post_mode == 1:
            body = re.sub(r'<a\sname=".*?"></a>\n', '', body)

        # 2 写入文件
        if output_dir is not None:
            title2 = refinepath(data['title'])
            f = XlPath(output_dir) / f'{title2}.md'
            f.write_text(body)

        return body

    def update_doc(self, repo_id, doc_id, doc_data):
        """ 更新单篇文档的详细信息

        :param repo_id: 知识库的ID或Namespace
        :param doc_id: 文档的ID
        :param doc_data: 包含文档更新内容的字典
        :return: 更新后的文档的详细信息
        """
        repo_id = self.get_repo_id(repo_id)
        url = f"{self.base_url}/repos/{repo_id}/docs/{doc_id}"
        response = requests.put(url, json=doc_data, headers=self.headers)
        return response.json()

    def update_doc_from_url(self, url, doc_data, *, md_cvt=True, return_mode='json'):
        """ 从文档的URL中更新文档的详细信息

        :param url: 文档的URL
        :param str|json doc_data: 包含文档更新内容的字典
            可以直接传入要更新的新的md内容
            注意无论原始是body_html、body_lake，都是要上传到body字段
        :param str return_mode: 返回的是更新后文档的内容，不过好像有bug，这里返回的body存储的并不是md格式
            'md', 返回更新后文档的主体md内容
            'json', 为原始json结构

            不建议拿这个返回值，完全可以另外再重新取返回值，就是正常的md格式了
        :param bool md_cvt: 是否需要转换md格式
            默认的md文档格式直接放回语雀，是会丢失换行的，需要对代码块外的内容，执行\n替换
        :return: 更新后的文档的详细信息
        """
        # 1 基础配置
        repo_slug, doc_slug = url.split('/')[-2:]
        if isinstance(doc_data, str):
            doc_data = {'body': doc_data}

        # 2 格式转换
        if md_cvt:
            ne = NestEnv(doc_data['body']).search(r'^```[^\n]*\n(.+?)\n^```',
                                                  flags=re.MULTILINE | re.DOTALL).invert()
            doc_data['body'] = ne.replace('\n', '\n\n')

        # 3 提交更新文档
        res = self.update_doc(repo_slug, doc_slug, doc_data)

        # 4 拿到返回值
        if return_mode == 'md':
            return res['data']
        elif return_mode == 'json':
            return res['data']['body']

    def __3_内容操作(self):
        pass

    def read_tables_from_doc(self, url, header=0):
        """ 从文档中读取表格

        :param url: 文档的URL
        :return: 表格列表
        """
        res = self.get_doc_from_url(url, return_mode=0)
        tables = pd.read_html(res['body_html'], header=header)
        return tables


if __name__ == '__main__':
    pass
